/* rndbg.c - random background on X11 display

   "rndbg" creates a random black and white image and installs it as
   background. The image is designed to be "neutral but not lifeless".
   It consists of a rectangle that is replicated to fill up the whole screen.

   The size of the rectangle can be changed using command line arguments.
   As an option, the image can be written on standard  output (in XBM format)
   instead of installed as the root image.

   The program has only been tested on Sun sparc running SunOS 4.1.1.

   The algorithm for generating the random image is quite simple.
   Start with an empty (black) image. Then place white pixels on random
   positions such that no two adjacent pixels are white (each pixel has
   four adjacent pixels; borders are treated in a "wrap-around" fashion).
   This process is repeated until no more white pixels can be placed.

   The result is quite fascinating. You have to see it!

   Niclas Wiberg, October 1994
   Linkoping University, Sweden
   nicwi@isy.liu.se
*/

#include <stdio.h>
#include <malloc.h>
#include <X11/Xlib.h>
#include <math.h>                /* We use floor() to select random pixel */
#include <time.h>                /* We use time() to get random number seed */
extern void srand48(long seed);  /* Initialize random number generator */
extern double drand48();         /* Return random in [0..1[ */

/*
 * Global parameters
 */
char *progname;
char *xbmname = NULL;
int width = 256;
int height = 256;
char *display = NULL;

/*
 * The result image in bitmap format
 */
char *bitmap = NULL;

/*
 * Forward declarations
 */
void args(int argc, char **argv);
void usage();
void rndbm();
void install_picture();
void output_xbm();


int main(int argc, char **argv)
{
  progname = argv[0];
  args(argc, argv);
  rndbm();
  if (xbmname) output_xbm();
  else install_picture();
  return 0;
}

void usage()
{
  fprintf(stderr,
	  "usage: %s [-xbm <name> | -display <display>] [width [height]]\n",
	  progname);
  exit(1);
}

void args(int argc, char **argv)
{
  int i;
  if (argc == 1) return;
  if (! strcmp(argv[1], "-xbm")) {
    if (argc < 3) usage();
    xbmname = argv[2];
    i = 3;
  } else  if (! strcmp(argv[1], "-display")) {
    if (argc < 3) usage();
    display = argv[2];
    i = 3;
  } else i = 1;
  if (i < argc) if (! sscanf(argv[i++], "%u", &width)) usage();
  height = width;
  if (i < argc) if (! sscanf(argv[i++], "%u", &height)) usage();
  if (i < argc) usage();
}

/*
 * install_picture()
 *
 * Install the result image as background pixmap on root window.
 */
void install_picture()
{
  Display *dpy;
  int screen;
  Window root;
  Pixmap pxm;
  
  dpy = XOpenDisplay(display);
  if (! dpy) {
    fprintf(stderr, "%s: can't open display %s\n", progname, display);
    exit(1);
  }
  screen = DefaultScreen(dpy);
  root = RootWindow(dpy, screen);
  pxm = XCreatePixmapFromBitmapData(dpy, root, bitmap, width, height,
				    BlackPixel(dpy, screen),
				    WhitePixel(dpy, screen),
				    DefaultDepth(dpy, screen));
  XSetWindowBackgroundPixmap(dpy, root, pxm);
  XFreePixmap(dpy, pxm);
  XClearWindow(dpy, root);
  XCloseDisplay(dpy);
}

/*
 * output_xbm()
 *
 * Write out the resulting bitmap in xbm format on standard output.
 */
void output_xbm()
{
  int i, n;
  printf("#define %s_width %d\n", xbmname, width);
  printf("#define %s_height %d\n", xbmname, height);
  printf("static char %s_bits[] = {\n", xbmname);
  n = (width + 7) / 8 * height;
  for (i = 0; i < n; i++) {
    if (i % 12 == 0) printf("   ");
    printf("0x%02x", bitmap[i] & 0xff);
    if (i == n - 1) printf("};\n");
    else if (i % 12 == 11) printf(",\n");
    else printf(", ");
  }
}


/*
 * The algorithm
 *
 * Although the algorithm is simple in theory, we have to make it
 * slightly more complicated to make it fast. The problem is to quickly
 * select random pixels that are "free", i.e., that do not have any
 * white neighbors. We do this by maintaining two data strucures:
 * A 2d-array for the pixels (pix), and a list pointing to the free
 * pixels (freepix). Pixels whose color has not yet been determined
 * point back to their positions in the freepix array.
 */

typedef struct {
  short x, y;
} coord;

/* pix[x][y] is either BLACK, WHITE or an index into freepix */
#define BLACK -1
#define WHITE -2
int **pix;

/* freepix[i], for i=0..remain-1, points to the free pixels in pix */
int remain;
coord *freepix;

/*
 * init()
 *
 * Allocate memory.
 * Initialize the freepix list, with back pointers from pix[x][y].
 */
void init()
{
  int x, y, i, j, k;

  pix = (int **) malloc(width * sizeof(int *));
  if (! pix) {perror(progname); exit(1);}
  for (x = 0; x < width; x++) {
    pix[x] = (int *) malloc(height * sizeof(int));
    if (! pix[x]) {perror(progname); exit(1);}
  }

  freepix = (coord *) malloc(width * height * sizeof(coord));
  if (! freepix) {perror(progname); exit(1);}

  bitmap = (char *) malloc((width + 7) / 8 * height * sizeof(char));
  if (! bitmap) {perror(progname); exit(1);}
  
  /* Initialize the freepix list, with back pointers from pix[x][y] */
  k = 0;
  for (x = 0; x < width; x++) {
    for (y = 0; y < height; y++) {
      pix[x][y] = k;
      freepix[k].x = x;
      freepix[k].y = y;
      k++;
    }
  }
  remain = height * width;
}
  
/*
 * set(int x, int y, int value)
 *
 * Assign a value (BLACK or WHITE) to a pixel.
 * If the pixel was free before, we have to remove it from the freepix list.
 * In this case, we may have to reorder the list, since we want the free
 * pixels to be indexed as 0..remain-1.
 */
void set(int x, int y, int value)
{
  int this; /* Index in freepix */
  
  /* Ensure that we are inside rectangle (wrap-around at borders) */
  while (x < 0) x += width;
  while (x >= width) x -= width;
  while (y < 0) y += height;
  while (y >= height) y -= height;

  /* Skip rest if pixel was set before (not free) */
  if (pix[x][y] < 0) return;

  /* Get this pixel's index in freepix list */
  this = pix[x][y];

  /* If this pixel was not last in the freepix list, we replace it with
     the last pixel, to avoid a "hole" in the list */
  if (this != remain - 1) {
    coord last = freepix[remain - 1];
    freepix[this] = last;
    pix[last.x][last.y] = this;
  }

  /* Now set the value and decrease the number of remaining free pixels */
  pix[x][y] = value;
  remain--;
}

/*
 * makebm()
 *
 * Convert pix[x][y] to bitmap data format.
 */
void makebm()
{
  int i, j, x, y, k;
  i = 0;
  for (y = 0; y < height; y++) {
    j = 0; k = 0;
    for (x = 0; x < width; x++) {
      if (pix[x][y] == BLACK) k |= (1 << j);
      j++;
      if (j == 8) {
	bitmap[i++] = k;
	j = 0; k = 0;
      }
    }
    if (j > 0) {
      bitmap[i++] = k;
    }
  }
}

void rndbm()
{
  int k, x, y;
  
  srand48(time(0));
  init();

  /* Place white pixels until there are no more possibilities */
  while (remain > 0) {

    /* Select random using freelist */
    k = (int) floor(remain * drand48());
    x = freepix[k].x;
    y = freepix[k].y;

    /* Selected pixel becomes white, its neighbors black */
    set(x, y, WHITE);
    set(x - 1, y, BLACK);
    set(x + 1, y, BLACK);
    set(x, y - 1, BLACK);
    set(x, y + 1, BLACK);
  }

  makebm();
}
